#!/usr/bin/env python
# Anna Agents-Oriented Testing Setup Loader

# Requires PyYAML: pip install pyyaml

import os, sys, datetime
from argparse import ArgumentParser

# PyYAML
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


def _exit(message = None, rc=0):
  if (message): printMsg(message)
  printMsg("Exiting [rc={}]".format(rc))
  sys.exit(rc)


def timeFormat():
  return '%b.%d.%Y-%H.%M.%S'


def printMsg(message):
  print ("[{}] {}".format(datetime.datetime.now().strftime(timeFormat()), message))


class YamlConfigParser():
  """
  Yaml parser
  """
  def __init__(self, yaml_config_file):
    """
    Convert the yaml file into a Python object
    """
    self.data = None
    try:
      with open(yaml_config_file, 'r') as ss:
        self.data = load(ss, Loader=Loader)
        ss.close()

    except Exception:
     raise

  def getData(self):
    return self.data

  def show(self, options):
    output = dump(self.data, Dumper=Dumper)
    print(output)


def writeFileContent(filename, content):
  _file = open(filename, "w")
  _file.write(content)
  _file.close()


def getNodeValue(dictionary, key):

  value = dictionary[key]
  printMsg("  {}: {}".format(key, value))
  return value


def configure_agent_adml(node, description, application_id, dictionary, _type, origin_host, origin_realm, address):

  # Agent ADML configuration is basically generation of services:

  if (_type == "client"):
    stacks='''  <stack id="{application_id}" dictionary="{dictionary}" fixMode="Always" ignoreFlagsOnValidation="yes"/>'''.format(application_id=application_id, dictionary=dictionary)

    nodes='''  <node originHost="{origin_host}" applicationId="{application_id}" entity="{address}" cer="cer.{application_id}.xml" answersTimeout="300000" dumpLog="yes"/>'''.format(application_id=application_id, dictionary=dictionary, origin_host=origin_host, address=address)

    ce_path = agents_ADML_dir + "/cer.{}.xml".format(application_id)
    ce_content='''<message version="1" name="CER" application-id="0">
   <avp name="Origin-Host" data="{origin_host}"/>
   <avp name="Origin-Realm" data="{origin_realm}"/>
   <avp name="Host-IP-Address" data="1|127.0.0.1"/>
   <avp name="Vendor-Id" data="10415"/>
   <avp name="Product-Name" data="{description}"/>
   <avp name="Auth-Application-Id" data="{application_id}"/>
</message>'''.format(application_id=application_id, origin_host=origin_host, origin_realm=origin_realm, description=description)

  elif (_type == "server"):
    stacks='''  <stack id="{application_id}" dictionary="{dictionary}" fixMode="Never"/>'''.format(application_id=application_id, dictionary=dictionary)

    nodes='''  <node originHost="{origin_host}" applicationId="{application_id}" diameterServer="{address}" diameterServerSessions="10" cea="cea.{application_id}.xml" answersTimeout="300000" dumpLog="yes"/>'''.format(application_id=application_id, dictionary=dictionary, origin_host=origin_host, address=address)

    ce_path = agents_ADML_dir + "/cea.{}.xml".format(application_id)
    ce_content='''<message version="1" name="CEA" application-id="0">
   <avp name="Result-Code" data="2001" alias="DIAMETER_SUCCESS"/>
   <avp name="Origin-Host" data="{origin_host}"/>
   <avp name="Origin-Realm" data="{origin_realm}"/>
   <avp name="Host-IP-Address" data="1|127.0.0.1"/>
   <avp name="Vendor-Id" data="10415"/>
   <avp name="Product-Name" data="{description}"/>
</message>'''.format(origin_host=origin_host, origin_realm=origin_realm, description=description)

  else:
    raise Exception("Invalid ADML type '{}' (allowed: client|server)".format(_type))

  # Create cer/cea files:
  writeFileContent(ce_path, ce_content)


  return stacks, nodes


def configure_agent_kafka(node, description, topic):

  # Producer script:
  content = '''#!/bin/bash
# Positional arguments: json, delay, debug indicator
[ -z "$3" ] && echo "Usage: $0 <json file> <delay in milliseconds: 0 is no delay> <debug indicator: yes|no>" && exit 1

json=$1
delay=""
[ "$2" != "0" ] && delay="-w $2"
debug=""
[ "$3" = "yes" ] && debug="-d"

echo "[{description}] Kafka Producer Step"
cd $(dirname $0)
python3 Producer.py -n {node} -t {topic} -f $json $delay $debug 2>&1 | tee -a {node}-producer.sh.output
exit ${{PIPESTATUS[0]}}
'''.format(description=description, node=node, topic=topic)

  path = agents_KAFKA_dir + "/{}-producer.sh".format(node)
  writeFileContent(path, content)
  os.chmod(path, 0o755)
  #printMsg("Created procedure '{}'".format(path))

  # Consumer script:
  content = '''#!/bin/bash
# Positional arguments: json, offset, timeout, debug indicator
[ -z "$4" ] && echo "Usage: $0 <json file ('any' to skip matching)> <auto offset reset> <timeout in seconds: 0 is no timeout> <debug indicator: yes|no>" && exit 1

file=""
[ "$1" != "any" ] && file="-f $1"
offset=$2
timeout=""
[ "$3" != "0" ] && timeout="-w $3"
debug=""
[ "$4" = "yes" ] && debug="-d"

echo "[{description}] Kafka Consumer Step"
cd $(dirname $0)
python3 Consumer.py -n {node} -t {topic} $file -o "$offset" $timeout $debug 2>&1 | tee -a {node}-consumer.sh.output
exit ${{PIPESTATUS[0]}}
'''.format(description=description, node=node, topic=topic)

  # We don't use -m (match as regular expression) because some problems have been detected with big json files.

  path = agents_KAFKA_dir + "/{}-consumer.sh".format(node)
  writeFileContent(path, content)
  os.chmod(path, 0o755)
  #printMsg("Created procedure '{}'".format(path))

  # Admin script:
  content = '''#!/bin/bash
# Positional arguments: operation, debug indicator
[ -z "$2" ] && echo "Usage: $0 <operation: clean_topic (at the moment)> <debug indicator: yes|no>" && exit 1
operation=""
case $1 in
  clean_topic)
    operation="-r"
    ;;
esac
debug=""
[ "$2" = "yes" ] && debug="-d"

echo "[{description}] Kafka Administration Step"
cd $(dirname $0)
python3 Admin.py -n {node} -t {topic} $operation $debug 2>&1 | tee -a {node}-admin.sh.output
exit ${{PIPESTATUS[0]}}
'''.format(description=description, node=node, topic=topic)

  path = agents_KAFKA_dir + "/{}-admin.sh".format(node)
  writeFileContent(path, content)
  os.chmod(path, 0o755)
  #printMsg("Created procedure '{}'".format(path))


def configure_agent_kafkacpp(node, description, topic):

  # Consumer script:
  content = '''#!/bin/bash
# Positional arguments: json, offset, timeout, debug indicator
[ -z "$4" ] && echo "Usage: $0 <json file ('any' to skip matching)> <auto offset reset> <timeout in seconds: 0 is no timeout> <debug indicator: yes|no>" && exit 1

file=""
[ "$1" != "any" ] && file="-f $1"
offset=$2
timeout=""
[ "$3" != "0" ] && timeout="-w $3"
debug=""
[ "$4" = "yes" ] && debug="-d"

echo "[{description}] Kafkacpp Consumer Step"
cd $(dirname $0)
sh Consumer -n {node} -t {topic} $file -o "$offset" $timeout $debug 2>&1 | tee -a {node}-consumer.sh.output
exit ${{PIPESTATUS[0]}}
'''.format(description=description, node=node, topic=topic)

  # We don't use -m (match as regular expression) because some problems have been detected with big json files.

  path = agents_KAFKACPP_dir + "/{}-consumer.sh".format(node)
  writeFileContent(path, content)
  os.chmod(path, 0o755)
  #printMsg("Created procedure '{}'".format(path))


def configure_agent_httpmock(node, description, address):

  # Provisioning script:
  content = '''#!/bin/bash
# Positional arguments: json, method, uri
[ -z "$3" ] && echo "Usage: $0 <test id> <step id> <test path> <json file> <method: POST|PUT|GET|DELETE> <uri>" && exit 1

echo "[{description}] Http Mock Provision Step"
cd $(dirname $0)
sh Provision.sh $1 $2 $3 "{address}" "$4" "$5" "$6" 2>&1 | tee -a {node}-provision.sh.output
exit ${{PIPESTATUS[0]}}
'''.format(description=description, node=node, address=address)

  path = agents_HTTPMOCK_dir + "/{}-provision.sh".format(node)
  writeFileContent(path, content)
  os.chmod(path, 0o755)
  #printMsg("Created procedure '{}'".format(path))


def configure_agents(agents):

  stacks = ""
  nodes = ""

  for node in agents:

    if (node == "ADML"): raise Exception('Invalid node id: "ADML" is reserved')
    if (node == "KAFKA"): raise Exception('Invalid node id: "KAFKA" is reserved')
    if (node == "KAFKACPP"): raise Exception('Invalid node id: "KAFKACPP" is reserved')
    if (node == "HTTPMOCK"): raise Exception('Invalid node id: "HTTPMOCK" is reserved')

    printMsg("Node: {}".format(node))

    description = getNodeValue(agents[node], "description")
    template = getNodeValue(agents[node], "template")

    if (template == "ADML"):
      # Node file:
      path = agents_ADML_dir + "/{}.node".format(node)
      writeFileContent(path, description)

      application_id = getNodeValue(agents[node], "application_id")
      dictionary = getNodeValue(agents[node], "dictionary")
      _type = getNodeValue(agents[node], "type")
      origin_host = getNodeValue(agents[node], "origin_host")
      origin_realm = getNodeValue(agents[node], "origin_realm")
      address = getNodeValue(agents[node], "address")

      # Here stacks and nodes are accumulated. Consolidation of 'services.xml' is done at the end of iteration (*)
      s,n = configure_agent_adml(node, description, application_id, dictionary, _type, origin_host, origin_realm, address)
      stacks += s
      stacks += "\n"
      nodes += n
      nodes += "\n"

    if (template == "KAFKA"):
      # Node file:
      path = agents_KAFKA_dir + "/{}.node".format(node)
      writeFileContent(path, description)

      topic = getNodeValue(agents[node], "topic")

      configure_agent_kafka(node, description, topic)


    if (template == "KAFKACPP"):
      # Node file:
      path = agents_KAFKACPP_dir + "/{}.node".format(node)
      writeFileContent(path, description)

      topic = getNodeValue(agents[node], "topic")

      configure_agent_kafkacpp(node, description, topic)


    if (template == "HTTPMOCK"):
      # Node file:
      path = agents_HTTPMOCK_dir + "/{}.node".format(node)
      writeFileContent(path, description)

      address = getNodeValue(agents[node], "address")

      configure_agent_httpmock(node, description, address)


  #(*) Consolidate services.xml
  path = agents_ADML_dir + "/services.xml"

  content='''<services>
{}
{}
</services>'''.format(stacks, nodes)

  writeFileContent(path, content)



def parse_arguments():

  parser = ArgumentParser(description='Anna Agents-Oriented Testing Setup Loader')
  parser.add_argument('-f', '--file', help='Agents yaml configuration file', required=True)


  arguments = parser.parse_args()

  return arguments


#####################
#      M A I N      #
#####################

if __name__ == "__main__":

  # Agents:
  abspath = os.path.abspath(__file__)
  dname = os.path.dirname(abspath)
  agents_ADML_dir = dname + "/agents/ADML"
  agents_KAFKA_dir = dname + "/agents/KAFKA"
  agents_KAFKACPP_dir = dname + "/agents/KAFKACPP"
  agents_HTTPMOCK_dir = dname + "/agents/HTTPMOCK"

  arguments = parse_arguments()

  afile = arguments.file

  try:
    agents = YamlConfigParser(arguments.file)

    configure_agents(agents.getData())
    #start_agents(agents)

    _exit()

  except Exception as e:
    _exit(e, 1)
