// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_oam_Configuration_hpp
#define anna_core_oam_Configuration_hpp


#include <anna/core/Singleton.hpp>
#include <anna/core/RuntimeException.hpp>

// STL
#include <string>
#include <vector>

namespace anna {
namespace xml {
class Node;
}
}


namespace anna {

namespace oam {


/**
* Class used for general OAM configuration
*/
class Configuration : public anna::Singleton <Configuration> {

  // Post-configuration for the final alarm text used for context data over each module (functionality, logic data, etc.):
  typedef std::vector<std::string> string_vector_t;
  string_vector_t a_alarm_text_preffix_components;
  string_vector_t a_alarm_text_suffix_components;
  bool a_alarms_preffix_enabled;   // Show general alarm preffix
  bool a_alarms_suffix_enabled;    // Show general alarm suffix


  // Delimiters (zone separator for preffix/text/suffix, and left/right/separator for preffix/suffix):
  char a_alarm_text_delimiter_zS;
  std::string a_alarm_text_delimiter_psL;
  std::string a_alarm_text_delimiter_psS;
  std::string a_alarm_text_delimiter_psR;

  const string_vector_t * getAlarmPreffixComponents() const throw() { return (a_alarms_preffix_enabled ? &a_alarm_text_preffix_components : NULL); }
  const string_vector_t * getAlarmSuffixComponents() const throw() { return (a_alarms_suffix_enabled ? &a_alarm_text_suffix_components : NULL); }

  void getAlarmTextDelimiters(char & zS, std::string & psL, std::string & psS, std::string & psR) const throw();


  Configuration(); // private constructor

  friend class anna::Singleton <Configuration>;
  friend class Module;

public:

  /**
  * Final detailed alarm text is composed by three parts: preffix, database text and suffix.
  * To separate these zones, we configure the zone separator (space character by default).
  * Within preffix/suffix we can configure left, right and separator delimiters (by default,
  * we use square brackets and pipe):
  *
  * <pre>
  *      ------- preffix -------        ------- suffix -------
  *     [ pc1 | pc2 | ... | pcM ] text [ sc1 | sc2 | ... | scN ]
  *
  *     Then, pc[i]/sc[i] are the preffix/suffix components.
  *     Usually, 'pc1' is the application facility name.
  *     Dashes could be used instead of pipe '|' with nice result.
  *
  * </pre>
  *
  * @param zS Zone separator delimiter
  * @param psL Preffix/Suffix left delimiter
  * @param psS Preffix/Suffix separator delimiter
  * @param psR Preffix/Suffix right delimiter
  */
  void setAlarmTextDelimiters(const char zS, const std::string & psL, const std::string & psS, const std::string & psR) throw();


  /**
  * Optional preffix definition to modify dinamically the alarm texts at context.
  * This affects to all the management modules. It is usually used on application
  * start tipically with the facility name, although this information is already
  * shown at the EG component but easier to identify through that detailed text.
  * Note that preffix components string should be multilanguage texts.
  *
  * @param components Preffix components for alarm text
  */
  void setAlarmPreffixComponents(const std::vector<std::string> & components) throw() { a_alarm_text_preffix_components = components; }


  /**
  * Optional suffix definition to modify dinamically the alarm texts at context.
  * This affects to all the management modules. It is usually used by application
  * libraries tipically with the oam module name. However this information could be
  * only useful for developers, and usually would be set by each oam module.
  * Note that suffix components string should be multilanguage texts.
  *
  * @param components Suffix components for alarm text
  */
  void setAlarmSuffixComponents(const std::vector<std::string> & components) throw() { a_alarm_text_suffix_components = components; }


  /**
  * Show general alarm preffix (enabled by default at constructor).
  */
  void enableAlarmsPreffix() throw();


  /**
  * Show general alarm suffix (enabled by default at constructor).
  */
  void enableAlarmsSuffix() throw();


  /**
  * Hide general alarm preffix (enabled by default at constructor).
  */
  void disableAlarmsPreffix() throw();


  /**
  * Hide general alarm suffix (enabled by default at constructor).
  */
  void disableAlarmsSuffix() throw();

  /**
  * Class XML representation
  *
  * @return XML with class content
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();
};

}
}

#endif

