// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_ClientSession_hpp
#define anna_diameter_comm_ClientSession_hpp


// STL
#include <string>

#include <anna/core/util/SortedVector.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/core/RuntimeException.hpp>

#include <anna/diameter.comm/Session.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/ClientSessionReceiver.hpp>
#include <anna/diameter.comm/ReceiverFactoryImpl.hpp>


namespace anna {
class DataBlock;
namespace timex {
class Engine;
}

namespace comm {
class Server;
}
}


namespace anna {

namespace diameter {

namespace comm {

class Server;


/**
   Modela la conexion realizada contra un servidor diameter.
*/
class ClientSession : public Session {

  // Helper:
  static std::string getKey(const std::string & addr, int port, int socketId) throw() {
    return (anna::functions::asString("%s:%d|%d", addr.c_str(), port, socketId));
  }

  bool a_hidden; // hide resource for restricted delivery over servers/entities


public:

  ClientSession();


  /* virtual */void initialize() throw();

  /**
   * Default watchdog period for the diameter client-session health.
   */
  static const anna::Millisecond DefaultWatchdogPeriod;

  /**
     Client session key: <address ip or hostname>:<remote port>|<socket id: 0..N-1>
  */
  std::string getKey() const throw() { return ClientSession::getKey(getAddress(), getPort(), getSocketId()); }

  /**
     Diameter server address (IPv4 or hostname)
     \return Diameter server address
  */
  /* virtual */const std::string& getAddress() const throw();

  /**
     Diameter server listen port
     \return Diameter server listen port
  */
  /* virtual */int getPort() const throw();

  /**
     Diameter parent server.
     \return Diameter parent server.
  */
  const Server *getParent() const throw() { return a_parent; }

  /**
     Diameter server created at diameter::comm::Engine::createClientSession.
     \return Diameter server
  */
  anna::comm::Server * getServer() throw() { return a_server; }

  /**
     Disables server resource (avoid the use of the server)
  */
  void disable() throw() { a_server->disable(); }

  /**
     Sets auto recovery indicator. When a connection is lost, by default it will be recovered automatically.
     \param autoRecovery Auto recovery indicator. TRue by default.
  */
  void setAutoRecovery(bool autoRecovery = true) throw() { a_autoRecovery = autoRecovery; a_server->setAutoRecovery(autoRecovery); }

  /**
     Gets the auto recovery indicator for the client connection (client-session).

     @return Auto recovery indicator.
  */
  bool getAutoRecovery() const throw() { return a_autoRecovery; }

  /**
     Sets the milliseconds wait to achieve a client connection to server by mean connect primitive.
     This allow to perform specific configurations (some servers could be slower than others).
     Changes will be taken into account on the next connect operation.

     \param maxConnectionDelay Milliseconds wait to get connection
  */
  void setMaxConnectionDelay(const anna::Millisecond & maxConnectionDelay) throw() { a_server->setMaxConnectionDelay(maxConnectionDelay); }

  /**
     Gets the milliseconds wait to achieve a client connection to server by mean connect primitive.

     \return Milliseconds wait to get connection
  */
  const anna::Millisecond & getMaxConnectionDelay() throw() { return a_server->getMaxConnectionDelay(); }

  /**
   * Sets CER and DWR diameter messages to be used over created client-sessions
   *
   * @param cer Capabilities-Exchange-Request message (encoded) for the client-sessions bind.
   * @param dwr Device-Watchdog-Request message (encoded) for the client-sessions keep-alive.
   */
  void setCERandDWR(const anna::DataBlock & cer, const anna::DataBlock & dwr) throw(anna::RuntimeException);

  // Internal
  void bind() throw(anna::RuntimeException);

  /* virtual */const Response* send(const Message* message) throw(anna::RuntimeException);
  /* virtual */bool unbind(bool forceDisconnect /* se usa en timer, para el actionTimer del tipo SessionUnbind, etc. */ = false) throw(anna::RuntimeException); // returns true if done at call time (no pendings or ignore pendings, except Disconnecting state by mean DPR/DPA)

  /**
     Deny resource for delivery restriction
  */
  void hide() throw() { a_hidden = true; }

  /**
     Allow resource for delivery permission
  */
  void show() throw() { a_hidden = false; }

  /**
     Returns true when client session resource is hidden for application messages delivery
  */
  bool hidden() const throw() { return a_hidden; }

  /**
     Returns true when client session resource is shown for application messages delivery
  */
  bool shown() const throw() { return !a_hidden; }


  /**
     Class string representation
     \return String with relevant information for this instance.
  */
  /* virtual */std::string asString() const throw();


  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  /* virtual */anna::xml::Node* asXML(anna::xml::Node* parent) const throw();


private:

  // Receiver factory
  ReceiverFactoryImpl<ClientSession, ClientSessionReceiver> a_receiverFactory;

  // Parent information
  Server *a_parent;

  // ClientSession messages:
  Message a_cer;
  Message a_dwr;

  // Connectivity
  bool a_autoRecovery;

  // Server
  anna::comm::Server *a_server;

  // Watchdog control:
  struct WatchdogState {
    enum _v {
      TimerStopped,           // Until CEA (bound state), timer is stopped
      WaitingTimerExpiration, // DWA has been received and we wait for next expiration to send DWR
      WaitingDWA              // DWR has been sent, but DWA hasn't been received yet
    };
  };
  WatchdogState::_v a_watchdogState;
  void setWatchdogState(WatchdogState::_v wState) throw();

  /* virtual */void expire(anna::timex::Engine *timeController) throw(anna::RuntimeException);
  void setWatchdogPeriod(const anna::Millisecond & watchdogPeriod) throw();

  /*virtual*/ void timerStopped() throw();
  /*virtual*/ void timerStarted() throw();


  // Activity:
  /* virtual */void updateIncomingActivityTime() throw();
  /* virtual */void updateOutgoingActivityTime() throw();
  void countSendings(const diameter::CommandId & cid, unsigned int aid, bool ok) throw();

  // Handlers:
  /**
     Handler about event break connection from diameter server over this client-session.

     When notified, ANNA.diameter.comm generates an diameter::comm::ClientSession::eventResponse for every request with pending answers.
  */
  void eventPeerShutdown() throw();

  /**
     Handler about a request retransmission over the session.

     \param request Message retransmitted
  */
  void eventRequestRetransmission(Message *request) throw();

  /**
     Handler for diameter server (client-session) responses

     \param response Answer container object for corresponding diameter request
  */
  void eventResponse(const Response& response) throw(anna::RuntimeException);

  /**
     Handler for diameter server (client-session) requests

     \param request Request data block object for corresponding diameter reception
  */
  void eventRequest(const anna::DataBlock& request) throw(anna::RuntimeException);
  //void eventRequest(const Message& request) throw(anna::RuntimeException);

  /**
     Handler for diameter server (client-session) responses out of context

     \param response Answer data block object without context match
  */
  void eventUnknownResponse(const anna::DataBlock& response) throw(anna::RuntimeException);

  /**
     Handler for diameter server (client-session) Disconnect-Peer-Answer messages

     \param response Answer data block object without context match
  */
  void eventDPA(const anna::DataBlock& response) throw(anna::RuntimeException);



  /**
  * Handlers for receptions
  */
  /* virtual */void receive(const anna::comm::Message& message) throw(anna::RuntimeException);
  /* virtual */void finalize() throw();
  void recover() throw();

  /* virtual */void expireResponse(Response*) throw();
  /* virtual */void setState(State::_v state) throw();

  void sendDWAToServer(const anna::DataBlock& dwrDB) throw(anna::RuntimeException); // non-usual behaviour, but DWR could be received from server

  // helpers
  static const char* asText(const WatchdogState::_v) throw();


  friend class Server;
  friend class Engine;
  friend class ClientSessionReceiver;
};

}
}
}

#endif

