// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_LocalServer_hpp
#define anna_diameter_comm_LocalServer_hpp


#include <anna/core/RuntimeException.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/core/util/Recycler.hpp>

// STL
#include <string>
#include <vector>
#include <map>

// Project
#include <anna/config/defines.hpp>
#include <anna/diameter.comm/ServerSession.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter.comm/MessageStatistics.hpp>


namespace anna {
namespace xml {
class Node;
}
namespace comm {
class ClientSocket;
//class LocalConnection;
}
class DataBlock;
}

namespace anna {

namespace diameter {

namespace comm {

class Engine;
class Response;
class ServerSocket;
class Message;


/**
   Diameter server socket
*/
class LocalServer {

  // main
  socket_t a_key;
  std::string a_description;
  int a_maxConnections;
  int a_currentConnections; // deberia coincidir en todo momento con el numero de local connections del server socket
  anna::Millisecond a_allowedInactivityTime;
  ServerSocket *a_serverSocket;
  int a_category;
  bool a_lock;

  // Engine
  Engine *a_engine;

  // Statistics
  MessageStatistics a_messageStatistics;
  void initializeStatisticResources() throw();
  void resetStatistics() throw();

//   void eraseServerSession(const anna::comm::ClientSocket& clientSocket) throw();
//   void eraseServerSession(const serverSession_iterator &it) throw();
  void lostConnection() throw();
  void newConnection() throw(anna::RuntimeException);

  // Activity
  anna::Millisecond a_lastIncomingActivityTime;   // last unix timestamp (in milliseconds) when message reception was managed over this entity
  anna::Millisecond a_lastOutgoingActivityTime;   // last unix timestamp (in milliseconds) when message sending was managed over this entity
  void updateIncomingActivityTime() throw();
  void updateOutgoingActivityTime() throw();

  // Availability
  bool a_available; // any of the server-sessions must be bound
  void availabilityLost() throw();
  void availabilityRecovered() throw();
  bool refreshAvailability() throw(); // return true if change

  void attach() throw(); // attach server socket to the communicator
  void attachPlanning() throw(); // used when attach fails (socket already in use, etc.)

  typedef int serverSession_key;
  serverSession_key getServerSessionKey(const anna::comm::ClientSocket&) const throw(); // hash for Client Socket INetAddress serialization
  typedef std::map <serverSession_key, ServerSession*> serverSession_container;
  typedef serverSession_container::value_type serverSession_value_type;
  typedef serverSession_container::iterator serverSession_iterator;
  typedef serverSession_container::const_iterator const_serverSession_iterator;
  serverSession_container a_serverSessions;
  anna::Recycler<ServerSession> a_serverSessionsRecycler;
  serverSession_iterator serverSession_find(const serverSession_key&) throw();
  serverSession_iterator serverSession_begin() throw() { return a_serverSessions.begin(); }
  serverSession_iterator serverSession_end() throw() { return a_serverSessions.end(); }
  static ServerSession* serverSession(serverSession_iterator ii) throw() { return ii->second; }
  const_serverSession_iterator serverSession_begin() const throw() { return a_serverSessions.begin(); }
  const_serverSession_iterator serverSession_end() const throw() { return a_serverSessions.end(); }
  static const ServerSession* serverSession(const_serverSession_iterator ii) throw() { return ii->second; }

  // INTERNAL CREATORS AND CLOSE METHODS
  ServerSession *createServerSession(const anna::comm::ClientSocket&) throw(anna::RuntimeException);
  void closeServerSession(ServerSession*) throw(anna::RuntimeException);

  // INTERNAL ALLOCATORS
  ServerSession* allocateServerSession() throw();
  void releaseServerSession(ServerSession*) throw();

  // Auxiliary
  serverSession_iterator a_deliveryIterator;
  ServerSession *a_lastUsedResource;

public:

  /** Constructor */
  LocalServer();

  /** Destructor */
  ~LocalServer() { close(); }


  // setters

  /**
  * Sets the local server key
  * @param LocalServer key
  */
  void setKey(const socket_t &key) throw() { a_key = key; }

  /**
     Sets the server socket optional description

     @param description Server socket description
  */
  void setDescription(const std::string description) throw() { a_description = description; }

  /**
     Set timeout to consider failed a request.
     \param v Requests class code.
     \param millisecond Milliseconds wait before considering the requests failed.

     Timers are internally managed and automatically activated.
  */
  void setClassCodeTimeout(const ClassCode::_v v, const anna::Millisecond & millisecond) throw();

  /**
     Sets the server socket optional category

     @param description Server socket category
  */
  void setCategory(int category) throw() { a_category = category; }

  /**
     Sets the maximum supported connections.
     If provided value is negative or lesser than the number of current connections, an exception will be launched.
     If all the connections was established, a new maximum will open the listen port.
     when margin is zeroed (maximum configured is equal to current connections), listen port will be closed.

     @param maxConnections Number of maximum connections allowed
  */
  void setMaxConnections(int maxConnections) throw(anna::RuntimeException);

  /**
     Sets the maximum allowed inactivity time on server sessions born over the local server before being reset.
     Communication engine assign a default value of 90000 msecs.

     @param allowedInactivityTime Inactivity time allowed
  */
  void setAllowedInactivityTime(const anna::Millisecond & allowedInactivityTime) throw() { a_allowedInactivityTime = allowedInactivityTime; }

  /**
  * Sets the diameter::comm::Engine
  * @param e Diameter::comm::Engine
  */
  void setEngine(Engine *e) throw() { a_engine = e; }

  // getters

  /**
  * Gets the diameter::comm::Engine
  * @return Diameter::comm::Engine
  */
  Engine *getEngine() const throw() { return a_engine; }

  /**
  * Gets the local server key
  * @return LocalServer key
  */
  const socket_t & getKey() const throw() { return a_key; }

  /**
     Gets the number of maximum accepted connections that server socket is configured to handle
  */
  int getMaxConnections() const throw() { return a_maxConnections; }

  /**
     Gets the number of current connections being established through server socket
  */
  int getCurrentConnections() const throw() { return a_currentConnections; }

  /**
     Gets the maximum allowed inactivity time on server sessions born over the local server before being reset

     @return Inactivity time allowed
  */
  const anna::Millisecond & getAllowedInactivityTime() const throw() { return a_allowedInactivityTime; }

  /**
     Returns true when any of the server-sessions is Bound. False when all not-bound.
  */
  bool isAvailable() const throw() { return a_available; }

  // helpers

  /**
     Disables local server socket (listener) keeping current server sessions alive.
     Note that applications should not close the listen port directly to keep coherence (see #resetConnectionsMargin)

     @param lock Locks disabled state (make it permanent even if new connections margin is reached).
     Used during diameter agent isolation (lost of service, maintenance, etc.)
  */
  void disable(bool lock = false) throw(anna::RuntimeException);

  /** Enables local server socket (listener)

     @param unlock Unlocks permanent disabled states
  */
  void enable(bool unlock = false) throw(anna::RuntimeException);

  /**
     Gets the number of requests messages over-the-air.

     @return OTA messages.
  */
  int getOTARequests() const throw();

  /**
     Returns idle state (no pending answers).

     @return Idle state.
  */
  bool idle() const throw() { return (getOTARequests() == 0); }

  /**
     Close the local server means two things: close the server socket and close all the server sessions born
     from this local server freeing such server sessions resources.
  */
  void close() throw(anna::RuntimeException);

  /**
     Performs coherent server socket close procedure zeroing margin between current established connections and maximum allowed.
  */
  void resetConnectionsMargin() throw(anna::RuntimeException) { setMaxConnections(a_currentConnections); }


  /**
   * Returns server-session instance identified by client socket provided.
   *
   * \param clientSocket Client socket associated to the server session
   * \param emode Action when no client-session is found with provided parameters (Throw/Ignore).
   *
   * \return The server-session instance identified by client socket provided.
   *
   * \warning If no server-session found, an exception is launched by default.
   */
  ServerSession* findServerSession(const anna::comm::ClientSocket &clientSocket, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);

  /**
   * Returns server-session instance identified by socket id provided (hash over serialized client socket information).
   *
   * \param socketId Socket id which is key for the server session
   * \param emode Action when no client-session is found with provided parameters (Throw/Ignore).
   *
   * \return The server-session instance identified by client socket provided.
   *
   * \warning If no server-session found, an exception is launched by default.
   */
  ServerSession* findServerSession(int socketId, anna::Exception::Mode::_v emode = anna::Exception::Mode::Throw) throw(anna::RuntimeException);


  /**
     Sent a message to the client using a certain server-session provided or defined by #readSocketId if not.
     When the message is a request, a timer will be set automatically to control the response time.
     If expires, the ResultCode Timeout will be finally notified on #LocalServer::eventResponse. This
     timeout value will be configured at #setClassCodeTimeout.

     \param message Message sent.
     \param socketId Server session socket id INetAddress serialization. By default, #readSocketId is invoked to get the socket id used (which uses round-robin if not re-implemented)

     @return Boolean about success in send operation. True when any of the server sessions could send the message.
     False, when neither of the server sessions was available or fail to send the message. Broadcast try to send all over
     the resources in spite of any fail. If a specific socket id is provided, only this socket is used without trying any other
     and returning false if fails.
  */
  bool send(const Message*, int socketId = -1 /* default uses readSocketId() */) throw(anna::RuntimeException);
  bool send(const Message& message, int socketId = -1 /* default uses readSocketId() */) throw(anna::RuntimeException) { return send(&message, socketId); }

  /**
     Gets the last used resource (server session) during sending.
     Broadcast doesn't updates this information.
  */
  ServerSession *getLastUsedResource() const throw() { return (a_lastUsedResource); }

  /**
     Before sending a message over each local server, socketId could be specified to select
     which session within such server will manage the message.

     Default implementation performs round-robin (value '-1' for socketId) but any other kind of
     application could re-implement this method and change the behaviour.

     \param message Message which is being sent.

     @return Socket-id (hash over serialized client socket information). Value '-1' if round-robin is desired.
     If socket-id is unkonwn, send procedure will throw an exception.
  */
  virtual int readSocketId(const Message *message) const throw() { return -1; }

  /**
     Sent a message to all the server sessions.
     It is used, i.e., in Disconnect-Peer-Request procedure over a certain entity.

     \param message Message which is being sent.

     @return Returns true (success) only when broadcast is success over all the entity servers. If any server fails,
     then false is returned.
  */
  bool broadcast(const Message *message) throw(anna::RuntimeException);
  bool broadcast(const Message &message) throw(anna::RuntimeException) { return broadcast(&message); }


  /**
     Class string representation
     \return String with relevant information for this instance.
  */
  std::string asString() const throw();

  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

  // Statistics
  void updateProcessingTimeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw();
  void updateReceivedMessageSizeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw();
//  int getProcessingTimeStatisticConcept() const throw() { return a_processing_time__StatisticConceptId; }
//  int getReceivedMessageSizeStatisticConcept() const throw() { return a_received_message_size__StatisticConceptId; }

protected:

  // Handlers:
  /**
     Handler about event break connection from diameter client over this server-session.
     When notified, ANNA.diameter.comm generates an diameter::comm::ServerSession::eventResponse for every request with pending answers.
     Default implementation traces warning event
     \param serverSession ServerSession from which shutdown has been received
  */
  virtual void eventPeerShutdown(const ServerSession* serverSession) throw();

  /**
     Handler about a request retransmission over the server-session.
     Default implementation traces warning event
     \param serverSession ServerSession from which retransmission happened
     \param request Retransmitted request message
  */
  virtual void eventRequestRetransmission(const ServerSession* serverSession, Message *request) throw();

  /**
     Handler for diameter client responses

     \param response Answer container object for corresponding diameter request
  */
  virtual void eventResponse(const Response& response) throw(anna::RuntimeException) = 0;

  /**
     Handler for diameter client requests

     \param serverSession ServerSession from which request has been received
     \param request Request data block object for corresponding diameter reception
  */
  virtual void eventRequest(ServerSession* serverSession, const anna::DataBlock& request) throw(anna::RuntimeException) = 0;
  //void eventRequest(ServerSession* serverSession, const Message& request) throw(anna::RuntimeException);

  /**
     Handler for diameter client responses out of context

     \param serverSession ServerSession from which request has been received
     \param response Answer data block object without context match
  */
  virtual void eventUnknownResponse(ServerSession* serverSession, const anna::DataBlock& response) throw(anna::RuntimeException) = 0;

  /**
     Handler for diameter client Disconnect-Peer-Answer messages

     \param serverSession ServerSession from which request has been received
     \param response Answer data block object without context match
  */
  virtual void eventDPA(ServerSession* serverSession, const anna::DataBlock& response) throw(anna::RuntimeException) = 0;


  friend class anna::diameter::comm::Timer;
  friend class Engine;
  friend class ServerSocket;
  friend class ServerSession;
  friend class ServerSessionReceiver;
};

}
}
}

#endif
