// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_testing_TestManager_hpp
#define anna_testing_TestManager_hpp

// Project
#include <anna/config/defines.hpp>
#include <anna/core/util/Recycler.hpp>
#include <anna/core/Singleton.hpp>
#include <anna/timex/TimeEventObserver.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/testing/TestTimer.hpp>
#include <anna/testing/TestCase.hpp>


namespace anna {
  class Millisecond;

  namespace timex {
    class Engine;
  }
  namespace diameter {
    namespace comm {
      class ClientSession;
      class ServerSession;
    }
  }

namespace testing {


class TestClock;
class TestCase;
class TestCaseStep;


typedef std::map<unsigned int /* test case id */, TestCase*> test_pool_t;
typedef std::map<unsigned int /* test case id */, TestCase*>::const_iterator test_pool_it;
typedef std::map<unsigned int /* test case id */, TestCase*>::iterator test_pool_nc_it;


/**
   Timer Manager for testing system
*/
class TestManager : public anna::timex::TimeEventObserver, public anna::Singleton <TestManager> {

  // Statistics summary:
  class StatSummary {

    unsigned int a_initializedTcs;
    unsigned int a_inprogressTcs;
    unsigned int a_failedTcs;
    unsigned int a_sucessTcs;

  public:
    StatSummary() { clear(); }
    void newTCState(const TestCase::State::_v beginState, const TestCase::State::_v endState) throw();
    void clear() throw();
    unsigned int getInitializedCount() const throw() { return a_initializedTcs; }
    unsigned int getInProgressCount() const throw() { return a_inprogressTcs; }
    unsigned int getFailedCount() const throw() { return a_failedTcs; }
    unsigned int getSuccessCount() const throw() { return a_sucessTcs; }
    unsigned int getFinishedCount() const throw() { return a_sucessTcs + a_failedTcs; }
    unsigned int getTotal() const throw() { return (a_initializedTcs + a_inprogressTcs + a_failedTcs + a_sucessTcs); }

    anna::xml::Node* asXML(anna::xml::Node* parent) const throw();
  };



  typedef anna::Recycler <TestTimer> timer_container;

  anna::timex::Engine* a_timeController;

  // reports
  std::string a_reportsDirectory;
  bool a_dumpInitializedReports, a_dumpInProgressReports, a_dumpFailedReports, a_dumpSuccessReports;
  bool a_dumpHexMessages;
  bool a_dumpStdout;

  // Pool of test cases
  test_pool_t a_testPool;
  test_pool_it a_currentTestIt;
  int a_poolRepeats; // repeat pool N times
  int a_poolCycle; // current cycle, from 1 to N
  unsigned int a_inProgressLimit; // limit load to have this value
  bool a_autoResetHard; // automatic reset on next cycle (soft by default, false)

  // Test clock
  int a_synchronousAmount;
  TestClock *a_clock;
  bool tick() throw();
  bool nextTestCase() throw();

  // Test timers
  timer_container a_timers;

  // Test case identifiers: key1 (strong); key2 (alternative)
  std::map<std::string /* key */, TestCase*> a_key1TestCaseMap;
  std::map<std::string /* key */, TestCase*> a_key2TestCaseMap;

  // Diameter shall use: key1 -> sessionId, key2 -> subscriberId
  //

  StatSummary a_statSummary; // general statistics


  TestManager();
  TestManager(const TestManager&);
  virtual ~TestManager() {;}

  TestTimer* createTimer(TestCaseStep*, const anna::Millisecond &, const TestTimer::Type::_v type) throw(anna::RuntimeException);
  void cancelTimer(TestTimer*) throw();
  void release(anna::timex::TimeEvent*) throw();


  public:

    void registerKey1(const std::string &key, const TestCase *testCase) throw(anna::RuntimeException);
    void registerKey2(const std::string &key, const TestCase *testCase) throw(anna::RuntimeException);

    int tests() const throw() { return a_testPool.size(); }
    void setTimerController(anna::timex::Engine *engine) throw() {  a_timeController = engine; }

    void setReportsDirectory(const std::string &rd) throw() { a_reportsDirectory = rd; }
    const std::string &getReportsDirectory() const throw() { return a_reportsDirectory; }

    void setDumpHex(bool dh) throw() { a_dumpHexMessages = dh; }
    bool getDumpHex() const throw() { return a_dumpHexMessages; }

    void setDumpStdout(bool ds) throw() { a_dumpStdout = ds; }
    bool getDumpStdout() const throw() { return a_dumpStdout; }

    void setDumpInitializedReports(bool enable) throw() { a_dumpInitializedReports = enable; }
    void setDumpInProgressReports(bool enable) throw() { a_dumpInProgressReports = enable; }
    void setDumpFailedReports(bool enable) throw() { a_dumpFailedReports = enable; }
    void setDumpSuccessReports(bool enable) throw() { a_dumpSuccessReports = enable; }
    void setDumpAllReports(bool enable) throw() {
      setDumpInitializedReports(enable);
      setDumpInProgressReports(enable);
      setDumpFailedReports(enable);
      setDumpSuccessReports(enable);
    }

    bool getDumpInitializedReports() const throw() { return a_dumpInitializedReports; }
    bool getDumpInProgressReports() const throw() { return a_dumpInProgressReports; }
    bool getDumpFailedReports() const throw() { return a_dumpFailedReports; }
    bool getDumpSuccessReports() const throw() { return a_dumpSuccessReports; }

    // Helper to calculate time interval and synchronous amount of execution tests to guarantee the input rate (tests per second)
    //  through the time manager. The first call to this method will start the time trigger system and check for new test cases to be launched.
    bool configureTTPS(int testTicksPerSecond) throw();

    bool clearPool(std::string &result) throw();
    bool resetPool(bool hard /* hard reset includes in-progress test cases */) throw();
    void setPoolRepeats(int repeats) throw() { a_poolRepeats = repeats; }
    int getPoolRepeats() const throw() { return a_poolRepeats; }
    int getPoolCycle() const throw() { return a_poolCycle; }

    bool getAutoResetHard() const throw() { return a_autoResetHard; }
    void setAutoResetHard(bool hard = true) throw() { a_autoResetHard = hard; }


    unsigned int getInProgressCount() const throw() { return a_statSummary.getInProgressCount(); }
    unsigned int getInitializedCount() const throw() { return a_statSummary.getInitializedCount(); }
    unsigned int getFinishedCount() const throw() { return a_statSummary.getFinishedCount(); }
    unsigned int getInProgressLimit() const throw() { return a_inProgressLimit; }
    void setInProgressLimit(unsigned int limit) throw() { a_inProgressLimit = limit; } // -1 = UINT_MAX (no limit)

    bool gotoTestCase(unsigned int id) throw();
    bool runTestCase(unsigned int id) throw();
    TestCase *findTestCase(unsigned int id) const throw(); // id = -1 provides current test case triggered
    TestCase *getTestCase(unsigned int id, const std::string &description = "") throw(); // creates/reuses a test case

    // Diameter-specific
    TestCase *getDiameterTestCaseFromSessionId(const anna::DataBlock &message, std::string &sessionId) throw();
    TestCase *getDiameterTestCaseFromSubscriberId(const anna::DataBlock &message, std::string &subscriberId) throw();
    void receiveDiameterMessage(const anna::DataBlock &message, const anna::diameter::comm::ClientSession *clientSession) throw(anna::RuntimeException);
    void receiveDiameterMessage(const anna::DataBlock &message, const anna::diameter::comm::ServerSession *serverSession) throw(anna::RuntimeException);

    // Non-trasactional (kafka)
    //
    //

    bool execTestCases(int sync_amount) throw();

    anna::xml::Node* asXML(anna::xml::Node* parent) const throw();
    anna::xml::Node* junitAsXML(anna::xml::Node* parent) const throw();
    std::string asXMLString() const throw();
    std::string junitAsXMLString() const throw();
    std::string summaryCounts() const throw();
    std::string summaryStates() const throw();

    // stats
    void tcsStateStats(const TestCase::State::_v beginState, const TestCase::State::_v endState) throw() {
      a_statSummary.newTCState(beginState, endState);
    }


  friend class anna::Singleton <TestManager>;
  friend class TestStepTimeout; // createTimer
  friend class TestStepDelay; // createTimer
  friend class TestClock; // tick
};

}
}

#endif

